"""Main"""

from flask import Blueprint, render_template

from ._issues import get_all_issues, prioritize_issues

main = Blueprint("main", __name__)


@main.route("/")
def index():
    """Index"""

    issues = get_all_issues()
    issues = prioritize_issues(issues)

    return render_template("index.html", issues=issues)



"""App Init"""

import argparse
import logging
import sys
from importlib.metadata import version

import setuptools
from flask import Flask
from github import Github
from gitlab import Gitlab
from sassutils.wsgi import SassMiddleware

from ._auth import github_login, gitlab_login
from ._config import get_app_config

__version__ = version("todo-merger")

parser = argparse.ArgumentParser(
    description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter
)
parser.add_argument(
    "-c",
    "--config-file",
    help="Path to the app config file in a non-default location",
)
parser.add_argument("--version", action="version", version="%(prog)s " + __version__)


logging.basicConfig(level=logging.DEBUG)


def load_app_config(config_file: str) -> dict[str, tuple[str, Github | Gitlab]]:
    """Load the app config, handle service logins, and return objects"""
    app_config: dict[str, dict[str, str]] = get_app_config(config_file)
    service_objects: dict[str, tuple[str, Github | Gitlab]] = {}

    for name, cfg in app_config.items():
        service = cfg.get("service", "")
        url = cfg.get("url", "")
        token = cfg.get("token", "")

        if not service:
            logging.critical(
                "The config section %s has no 'service' defined, e.g. 'github' or 'gitlab'", name
            )
            sys.exit(1)

        if service == "gitlab" and not url:
            logging.critical(
                "The config section %s is a gitlab service but has no 'url' defined", name
            )
            sys.exit(1)

        if name in service_objects:
            logging.critical(
                "You have used the section name %s more than once. Please make them unique", name
            )
            sys.exit(1)

        if service == "github":
            loginobj = github_login(token)
        elif service == "gitlab":
            loginobj = gitlab_login(token, url)  # type: ignore
        else:
            logging.critical("The config section %s contains an unknown 'service'", name)
            sys.exit(1)

        service_objects[name] = (service, loginobj)

    return service_objects


# pylint: disable=import-outside-toplevel
def create_app(config_file: str):
    """Create Flask App"""
    app = Flask(__name__)
    # Reload templates
    app.jinja_env.auto_reload = True
    app.jinja_env.lstrip_blocks = False
    app.config["TEMPLATES_AUTO_RELOAD"] = True

    app.wsgi_app = SassMiddleware(  # type: ignore
        app.wsgi_app,
        {
            "todo_merger": {
                "sass_path": "static/sass",
                "css_path": "static/css",
                "wsgi_path": "/static/css",
                "strip_extension": False,
            }
        },
    )

    # Load app config and login to services (e.g. GitHub and GitLab)
    app.config["services"] = {}
    for name, service in load_app_config(config_file).items():
        app.config["services"][name] = service

    # blueprint for app
    from .main import main as main_blueprint  # pylint: disable=import-error

    app.register_blueprint(main_blueprint)

    return app


def main():
    """Main entry point for running the app."""
    args = parser.parse_args()
    app = create_app(config_file=args.config_file)
    app.run(port=8636)



"""Handling issues which come in from different sources"""

import logging
from dataclasses import dataclass, fields
from urllib.parse import urlparse

from flask import current_app
from github import Github
from gitlab import Gitlab


@dataclass
class IssueItem:  # pylint: disable=too-many-instance-attributes
    """Dataclass holding a single issue"""

    assignee_user: str = ""
    due_date: str = ""
    epic_title: str = ""
    milestone_title: str = ""
    ref: str = ""
    title: str = ""
    web_url: str = ""
    service: str = ""

    def import_values(self, **kwargs):
        """Import data from a dict"""
        for attr, value in kwargs.items():
            setattr(self, attr, value)


def gitlab_get_issues(gitlab: Gitlab) -> list[IssueItem]:
    """Get all issues assigned to authenticated user"""
    issues: list[IssueItem] = []
    for issue in gitlab.issues.list(
        assignee_username=gitlab.user.username, state="opened"  # type: ignore
    ):
        # See https://docs.gitlab.com/ee/api/issues.html
        d = IssueItem()
        d.import_values(
            # TODO: get all assignees, filter out oneself
            assignee_user=issue.assignee["username"] if issue.assignee else "",
            due_date=issue.due_date,
            epic_title=issue.epic["title"] if issue.epic else "",
            milestone_title=issue.milestone["title"] if issue.milestone else "",
            ref=issue.references["full"],
            title=issue.title,
            web_url=issue.web_url,
            service="gitlab",
        )
        issues.append(d)

    return issues


def _gh_url_to_ref(url: str):
    """Convert a GitHub issue URL to a ref"""
    url = urlparse(url).path
    url = url.strip("/")
    return url.replace("/issues/", "#")


def github_get_issues(github: Github) -> list[IssueItem]:
    """Get all issues assigned to authenticated user"""
    issues: list[IssueItem] = []
    for issue in github.get_user().get_issues():  # type: ignore
        # See https://docs.github.com/en/rest/issues/issues
        d = IssueItem()
        d.import_values(
            assignee_user=issue.assignee.login if issue.assignee else "",
            due_date="",
            epic_title="",
            milestone_title=issue.milestone.title if issue.milestone else "",
            ref=_gh_url_to_ref(issue.html_url),
            title=issue.title,
            web_url=issue.html_url,
            service="github",
        )
        issues.append(d)

    return issues


def get_all_issues() -> list[IssueItem]:
    """Get all issues from the supported services"""
    issues: list[IssueItem] = []
    for name, service in current_app.config["services"].items():
        if service[0] == "github":
            logging.info("Getting assigned GitHub issues for %s", name)
            issues.extend(github_get_issues(service[1]))
        elif service[0] == "gitlab":
            logging.info("Getting assigned GitLab issues for %s", name)
            issues.extend(gitlab_get_issues(service[1]))

    return issues


def _replace_none_with_empty_string(obj: IssueItem) -> IssueItem:
    """Replace None values of a dataclass with an empty string. Makes sorting
    easier"""
    for field in fields(obj):
        value = getattr(obj, field.name)
        if value is None:
            setattr(obj, field.name, "")

    return obj


def prioritize_issues(
    issues: list[IssueItem], sort_by: list[tuple[str, bool]] | None = None
) -> list[IssueItem]:
    """
    Sorts the list of IssueItem objects based on multiple criteria.

    :param issues: List of IssueItem objects to sort.
    :param sort_by: List of tuples where each tuple contains:
                    - field name to sort by as a string
                    - a boolean indicating whether to sort in reverse order
                      (True for descending, False for ascending)
    :return: Sorted list of IssueItem objects.
    """
    if sort_by is None:
        sort_by = [("due_date", False), ("milestone_title", True)]

    logging.info("Sort issues based on %s", sort_by)

    # Replace None with empty string in all tasks
    issues = [_replace_none_with_empty_string(task) for task in issues]

    def sort_key(issue: IssueItem) -> tuple:
        # Create a tuple of the field values to sort by, considering the reverse order
        key: list[tuple[int, str | None]] = []
        for field, reverse in sort_by:
            value: str = getattr(issue, field).lower()
            is_empty = value == ""
            # Place empty values at the end
            if is_empty:
                key.append((1, None))  # `1` indicates an empty value
            else:
                if reverse:
                    value = "".join(chr(255 - ord(char)) for char in value)
                key.append((0, value))  # `0` indicates a non-empty value
        return tuple(key)

    return sorted(issues, key=sort_key)
